import pandas as pd
from sklearn.feature_selection import mutual_info_regression

def get_mi_scores(features: pd.DataFrame, target: pd.DataFrame, discrete_features: list[bool]) -> pd.Series:
    """Calculate Mutual Information (MI) Scores for feature selection.

    Args:
        features (pd.DataFrame): The DataFrame containing the features.
        target (pd.DataFrame): The DataFrame containing the target variable.
        discrete_features (list[bool]): A list of Boolean values indicating whether each feature
            is discrete (True) or continuous (False).

    Returns:
        pd.Series: A Series containing the MI Scores for each feature, sorted in descending order.

    Source: https://www.kaggle.com/code/ryanholbrook/principal-component-analysis
    """
    
    mi_scores = mutual_info_regression(features, target, discrete_features=discrete_features)
    mi_scores = pd.Series(mi_scores, name="MI Scores", index=features.columns)
    mi_scores = mi_scores.sort_values(ascending=False)

    return mi_scores


import pandas as pd
import difflib  # used to compare strings
import re  # regular expressions

# ! --------------------------Column and Row Cleaning--------------------------

def remove_high_cardinality_cols(df: pd.DataFrame, threshold: int) -> pd.DataFrame:
    """Remove columns with a cardinality higher than the given threshold.

    Args:
        df (pd.DataFrame): Pandas DataFrame to update
        threshold (int): Threshold of cardinality
    """

    # Identify columns that exceed the threshold
    high_cardinality_cols = df.nunique()[df.nunique() > threshold].index

    # Print the dropped columns
    print(
        len(high_cardinality_cols),
        "features/columns dropped due to high cardinality:",
        high_cardinality_cols.values,
    )

    # Drop these columns from the dataframe
    return df.drop(high_cardinality_cols, axis=1)


def remove_duplicates(df: pd.DataFrame) -> pd.DataFrame:
    """Prints info about and removes duplicate columns and rows

    Args:
        df (pd.DataFrame): Incoming Pandas DataFrame

    Returns:
        pd.DataFrame: Pandas DataFrame with no duplicate rows/columns
    """

    df_copy = df.copy()

    # count and remove duplicate rows
    duplicate_rows = df_copy[df_copy.duplicated()]
    num_duplicate_rows = len(duplicate_rows)
    df_copy = df_copy.drop_duplicates()

    # count and remove duplicate columns
    duplicate_columns = df_copy.columns[df_copy.columns.duplicated()]
    num_duplicate_columns = len(duplicate_columns)
    df_copy = df_copy.loc[:, ~df_copy.columns.duplicated()]

    print(f"{num_duplicate_rows} rows removed: {duplicate_rows}")
    print(f"{num_duplicate_columns} columns removed: {duplicate_columns}")

    return df_copy


# ! --------------------------String/Object Cleaning---------------------------

def show_spelling_errors(
    df: pd.DataFrame, similarity_threshold: float, exclude_columns: list[str] = []
) -> None:
    """This prints all of the observations in a column that are similar above a threshold

    Args:
        df (pd.DataFrame): Pandas DataFrame
        similarity_threshold (float): Decimal of how similar of results we want to see (0.0-1.0)
        exclude_columns (list[str]): List of columns you want to exclude from spelling check
    """

    spelling_errors = {}

    # find potential spelling errors for object columns
    for column in df.select_dtypes(include="object"):
        if column not in exclude_columns:
            unique_values = df[column].dropna().unique()
            potential_errors = []

            for i, value1 in enumerate(unique_values):
                for value2 in unique_values[i + 1 :]:
                    similarity = difflib.SequenceMatcher(None, value1, value2).ratio()
                    if similarity > similarity_threshold:
                        potential_errors.append((value1, value2))

            if potential_errors:
                spelling_errors[column] = potential_errors

    # print the errors
    for column, errors in spelling_errors.items():
        print(f"Potential spelling errors in column '{column}':")
        for error in errors:
            print(f"- '{error[0]}' might be similar to '{error[1]}'")


def remove_strings_from_cols(
    df: pd.DataFrame, columns_to_clean: list[str], strings_to_remove: list[str]
) -> pd.DataFrame:
    """
    Removes specified strings from specified columns in a Pandas DataFrame and returns a copy of the DataFrame.

    Args:
        df (pd.DataFrame): The input DataFrame.
        columns_to_clean (list[str]): A list of column names to clean.
        strings_to_remove (list[str]): A list of strings to remove from the specified columns.

    Returns:
        pd.DataFrame: A new DataFrame with the specified strings removed from the specified columns, leaving the original DataFrame unchanged.
    """

    df_copy = df.copy()

    for col in columns_to_clean:
        for string in strings_to_remove:
            df_copy[col] = df_copy[col].apply(lambda x: re.sub(string, "", x).strip())

    return df_copy


def alpha_numeric_text_clean(text: str) -> str:
    """Removes any extra spaces or special characters from text.

    Args:
        text (str): Text that you want to filer

    Returns:
        str: The cleaned text
    """

    # remove anything other than letter, number, and spaces then make lowercase
    text = re.sub(r"[^a-zA-Z0-9\s]", "", text).lower()

    # remove extra spaces and leading/trailing spaces
    text = " ".join(text.split())

    return text


# ! -------------------------------NaN Cleaning--------------------------------

def remove_high_nan_cols(df: pd.DataFrame, threshold: int) -> pd.DataFrame:
    """Remove columns with a NaN count higher than the given threshold.

    Args:
        df (pd.DataFrame): Pandas DataFrame to update
        threshold (int): Threshold of NaN count
    """

    high_nan_cols = pd.isnull(df).sum()[pd.isnull(df).sum() > threshold].index

    print(len(high_nan_cols), "features/columns dropped:", high_nan_cols.values)

    return df.drop(high_nan_cols, axis=1)


def print_all_nan_counts(df: pd.DataFrame) -> None:
    """Prints the number of NaNs for each column of the DataFrame.

    Args:
        df (pd.DataFrame): Pandas DataFrame
    """

    nan_counts = df.isnull().sum().sort_values(ascending=False)
    print(f"NaN Counts:\n{nan_counts}")


def print_nan_cols(df: pd.DataFrame) -> None:
    """Prints the number of NaNs in columns that have NaNs.

    Args:
        df (pd.DataFrame): Pandas DataFrame
    """

    nan_counts = df.isnull().sum().sort_values(ascending=False)
    nan_counts = nan_counts[nan_counts > 0]
    print(f"NaN Counts:\n{nan_counts}")


def find_nan_columns(df: pd.DataFrame) -> pd.Index:
    """Returns the columns that have NaN values.

    Args:
        df (pd.DataFrame): Pandas DataFrame

    Returns:
        pd.Index: Pandas Index of NaN columns
    """

    nan_features = df.isnull().sum()
    non_zero_nans = nan_features[nan_features > 0]

    return non_zero_nans.index



# https://gaurav-adarshi.medium.com/logging-module-in-python-a3ed49257928

import logging
import colorlog

# Create a logger
logger = logging.getLogger(__name__)

# Create a ColoredFormatter with the desired format
formatter = colorlog.ColoredFormatter(
    "%(asctime)s %(log_color)s%(levelname)s%(reset)s: %(message)s",
    datefmt="%m-%d-%Y %H:%M:%S",
    reset=True,
    log_colors={
        'DEBUG': 'cyan',
        'INFO': 'green',
        'WARNING': 'yellow',
        'ERROR': 'red',
        'CRITICAL': 'white,bg_red',
    },
    secondary_log_colors={},
    style='%'
)

# Create a StreamHandler and set the formatter
console_handler = logging.StreamHandler()
console_handler.setFormatter(formatter)

# Add the console handler to the logger
logger.addHandler(console_handler)

# Set the logging level (e.g., DEBUG, INFO, WARNING, ERROR, CRITICAL)
logger.setLevel(logging.DEBUG)


logger.debug("This is a debug message.")
logger.info("This is an info message.")
logger.warning("This is a warning message.")
logger.error("This is an error message.")
logger.critical("This is a critical message.")


import yaml
import logging
import pandas as pd
import os

def summarize_file(df: pd.DataFrame, file_path: str) -> None:
    """Prints a summary of a data file including total rows, columns, and file size in MB.

    Args:
        df (pd.DataFrame): Pandas DataFrame
        file_path (str): File path
    """

    # filesize in mb
    file_size_bytes = os.path.getsize(file_path)
    file_size_mb = file_size_bytes / (1024 * 1024)

    # get dimensions
    total_rows = len(df)
    total_columns = len(df.columns)

    print(f"Total number of rows: {total_rows}")
    print(f"Total number of columns: {total_columns}")
    print(f"File size: {file_size_mb:.2f} MB")


def read_config_file(filepath: str) -> dict:
    """Reads a YAML file for data ingestion.

    Args:
        filepath (str): YAML file path

    Returns:
        dict: YAML data
    """

    with open(filepath, "r") as stream:
        try:
            return yaml.safe_load(stream)
        except yaml.YAMLError as exc:
            logging.error(exc)


def num_col_validation(df: pd.DataFrame, table_config: dict) -> bool:
    """Validates if the number of columns in the DataFrame matches the table configuration.

    Args:
        df (pd.DataFrame): Pandas DataFrame
        table_config (dict): Validation data

    Returns:
        bool: If number of columns match the validation data
    """

    if len(df.columns) == len(table_config["columns"]):
        return True
    else:
        return False


def col_header_val(df: pd.DataFrame, table_config: dict) -> bool:
    """Validates if the header names in the DataFrame match the table configuration.

    Args:
        df (pd.DataFrame): Pandas DataFrame
        table_config (dict): Validation data

    Returns:
        bool: If column headers match the validation data
    """

    # sort, strip leading and trailing spaces, and replace space with _
    df_columns = sorted([col.strip().lower().replace(" ", "_") for col in df.columns])
    yaml_columns = sorted(
        [col.strip().lower().replace(" ", "_") for col in table_config["columns"]]
    )

    if df_columns == yaml_columns:
        return True
    else:
        # find the mismatched columns
        mismatched_columns = set(df_columns) ^ set(yaml_columns)
        print(f"Mismatched columns: {list(mismatched_columns)}")
        return False


from setuptools import setup, find_packages

setup(
    name='zoomds',
    version='0.1',
    author='Matthew Iversen',
    author_email='iversenmatt7@gmail.com',
    description='A collection of data science and analysis functions for data cleaning, analysis, modeling, and visualization.',
    packages=find_packages(),
    install_requires=[
        'colorlog',
        'matplotlib',
        'numpy',
        'pandas',
        'scipy',
        'scikit-learn',
        'pyyaml',
    ],
    python_requires='>=3.6',
)
